#ifndef _KEYSPACEMANAGER_H
#define _KEYSPACEMANAGER_H
#include "settings.h" //curses settings, etc
#include "datatypes.h"
#include "keyspacechunk.h"
#include "keyspacechunkstatus.h"
#include "key.h"
#include "timepiece.h"
#include <map>
#include <list>
#include <iostream>
#include <fstream>

#ifdef HAVE_NCURSES
#include <ncurses.h>
#endif
using namespace std;

typedef map<Key, KeySpaceChunk> ChunkMapType;
class KeySpaceManager
{
	private:
	//the basics. These dont need to be saved in the restore function because the ctor will recompute them.
	u_int32_t KeySpaceSizeInBits;
	u_int32_t KeySpaceSizeInBytes; 
	u_int32_t Range;
	Key TotalChunks;    // TOTAL number of chunks in the entire keyspace. We use this to know when we're done.
	
	//only CheckedOutChunks needs to be saved. StragglerList will be repopulated after it is loaded.
	ChunkMapType	CheckedOutChunks;
	//FIX THIS, AND LOOKFORSTRAGGLERS TOMORROW
	list <KeySpaceChunk*> StragglerList;
	
	//important state stuff
	//this is also saved/restored
	Key *NextInitialKey;// value we start the next key chunk with.

	// ----------statistics and counters---------
	//Statistics for display
	//all these are saved/restored		
	bool CanComputeIntegerStats; //mathematically, if its > 2^32 we cant.
	u_int32_t IntTotalChunks; // lets us compute percent, estimated time, etc
	u_int32_t IntCheckedOutChunkCounter; //Counts how many chunks have been checked OUT ever
	u_int32_t IntCheckedInChunkCounter; //Counts how many chunks have been checked back IN
	
	//these are glorified counters. used for display. We cant necesarrily use the integers above
	//cause the numbers get to damn big/ so we use numbers for math and these keys for display
	//its not the  most pretty solution but it works. If i added math operators to the key
	//class it would be unnecesarry

	Key CheckedOutChunkCounter; //Counts how many have been checked OUT, not checked back in
	Key CheckedInChunkCounter;  //Counts how many have been checked IN, not checked back in
					 		    //When this == TotalChunks, we're all done.

	//we can compute this even if CanComputeIntegerStats = false becaues its unlikely 
	//to roll over 32 bits.
	u_int32_t IntExpiredCounter;		//How many keychunks we've re-assigned	


	//---time--- 
	//these are saved/restored
	TimePiece TotalComputeTime;
	TimePiece TotalRunningTime;	
	
	//-----Misc Settings------
	//none of these need to be saved/restored. The ctor will recompute them
	ofstream LogFile;
	bool Logging;
	bool Verbose;
	u_int32_t StragglerThreshold;
	
	//internal curses functions for drawing display
	#ifdef HAVE_NCURSES
	void DrawBg();
	void DrawFg();
	#endif	
	public:
	void LookForStragglers();	
	void IncrementStragglerThreshold(u_int32_t value=FIFTEEN_MINS);
	void DecrementStragglerThreshold(u_int32_t value=FIFTEEN_MINS);

	// the keysize were cracking and the size of chunks to hand out
	KeySpaceManager(u_int32_t KeySpaceSizeInBits, u_int32_t Range);
	bool SaveState(FILE *fp);
	bool RestoreState(FILE *fp);
	
	// returns an initial key to start with and a range to go through;
	bool getKeySpaceChunk(KeySpaceChunk &chunk);
	void submitCompletedKeySpaceChunk(Key iKey, u_int32_t session_id);
	//u_int32_t getChunksCheckedIn(); 
	u_int32_t getChunksCheckedOut(); 
	bool StillMoreChunks();
	//display functions	
	void UpdateDisplay(u_int32_t num_secs=0);
	void PrintStatistics();
	void RedrawDisplay();
	~KeySpaceManager();
};


#endif
