#include "keyspacemanager.h"
#include "datatypes.h"
#include "typedefs.h"
#include "arguments.h"
#include "unistd.h"
#include <math.h>

extern ArgumentsType GlobalArguments;
//
//keyspacesize and chunksize are in bits

//KeySpaceManager :: KeySpaceManager(u_int32_t _keyspacesizeisbits, u_int32_t _range)
KeySpaceManager :: KeySpaceManager(u_int32_t _keyspacesizeinbits, u_int32_t _rangesizeinbits)
{
	char temp[1024];
	KeySpaceSizeInBits = _keyspacesizeinbits;
	Range = (u_int32_t)pow((double)2, (double) _rangesizeinbits);
	printf("rangesizeinbits is %d\n", _rangesizeinbits);
	printf("RAnge is %d\n", Range);

	//copmute wheter or not we can compute the total chunks
	if (_keyspacesizeinbits - _rangesizeinbits <= 32)
	{
		CanComputeIntegerStats = true;
		IntCheckedInChunkCounter = 0;
		IntCheckedOutChunkCounter = 0;
		IntTotalChunks = (u_int32_t) pow((double)2, (double) _keyspacesizeinbits - _rangesizeinbits);
		cout << "IntTotalChunks: " << IntTotalChunks << endl;
	}
	else
	{
		CanComputeIntegerStats = false;
		IntTotalChunks = 0;
		cout << "Cannot compute pow(2,  (keyspacesizeinbits - chunksizeinbits)\n";
		cout << "becase keyspacesize in bits - chunk size in bits = " << _keyspacesizeinbits - _rangesizeinbits;
		cout << " Which is > 32\n";
		cout << "This makes computing certain statistics for the display impossible.\n";
		sleep(1);
	}
	//
	//regardless of whether we can compute the total chunks  as an integer
	//we can compute it as a Key and display it atleast.
	//we also use this key to know when we've assigned all the chunks out
	u_int32_t SizeOfTotalChunkCounterInBits = _keyspacesizeinbits - _rangesizeinbits;
	u_int32_t SizeOfTotalChunkCounterInBytes = 0;

	
	SizeOfTotalChunkCounterInBytes = (SizeOfTotalChunkCounterInBits / 8) + 1;
	
	/*
	if (SizeOfTotalChunkCounterInBits % 8 == 0)
	{
		cout << "Clean byte alignment\n";
		SizeOfTotalChunkCounterInBytes = SizeOfTotalChunkCounterInBits / 8;
		cout << "We need 1 more byte though\n";
		SizeOfTotalChunkCounterInBytes++;
	}
	else 
	{
		cout << "Unclean byte alignment\n";
		SizeOfTotalChunkCounterInBytes = (SizeOfTotalChunkCounterInBits / 8) + 1;
	}
	*/
	TotalChunks.setSize(SizeOfTotalChunkCounterInBytes);
	CheckedOutChunkCounter.setSize(SizeOfTotalChunkCounterInBytes);
	CheckedInChunkCounter.setSize(SizeOfTotalChunkCounterInBytes);

#ifdef JC_DEBUG
	cout << "Size of TotalChunkCounter is " << SizeOfTotalChunkCounterInBytes;
	cout << "Is that big enogh to represent " << _keyspacesizeinbits - _rangesizeinbits  << "bits?" << " ?\n";
#endif
	//Now that we know the size in bytes/bits of our Counters we need to set the TotalChunks
	//to its Max. Unfortunately since the Key class measures things in bytes we need to do some surgery
	{
		u_int32_t num_whole_bytes = (_keyspacesizeinbits - _rangesizeinbits)/ 8;
		u_int32_t num_extra_bits = (_keyspacesizeinbits - _rangesizeinbits) %8;
		
		unsigned char *p;
		unsigned char tmp;
		p = TotalChunks.value;
		//p += num_whole_bytes;
		tmp = (unsigned char) pow((double)2, (double)num_extra_bits);

		cout << "Before: " << TotalChunks  << endl;
		printf("2^ %d is %x\n", num_extra_bits, tmp);
		cout << "There are " << num_whole_bytes << "whole bytes and " << num_extra_bits << "extra bits in "
			<< (_keyspacesizeinbits - _rangesizeinbits) << endl;

		memset(p , tmp, 1);

		cout << "After: " << TotalChunks << endl;
		cout << "that value should be 2^" << _keyspacesizeinbits - _rangesizeinbits << endl;
		cout << "Which is " << IntTotalChunks << endl;
		}	
	
	
	
	
	if (KeySpaceSizeInBits % 8 != 0)
	{
		fprintf(stderr, "KeySpaceManager:: Must pass keyspacesize in bits!\n");
		exit(0);
	}
	else
		KeySpaceSizeInBytes = KeySpaceSizeInBits / 8;

	
	if (KeySpaceSizeInBits > MAX_KEY_SIZE_IN_BITS)
	{
		snprintf(temp, sizeof(temp), 
						"Cant create KeySpaceManager with keyspacesize( %d) > MAX_KEY_SIZE_IN_BITS (%d) \n",
						_keyspacesizeinbits, MAX_KEY_SIZE_IN_BITS);
		fprintf(stderr,temp);
		fprintf(stderr,"edit libjcwepcrack/config.h and recompile\n");
		fprintf(stderr,"If you use a key > 128 bits the nice ncurses display will need some touching up\n");
		exit(0);
	}


	StragglerThreshold = GlobalArguments.straggler_threshold;
	if (GlobalArguments.verbose)
			Verbose = true;
	else
			Verbose = false;

	if (GlobalArguments.logging)
	{
		LogFile.open(GlobalArguments.LogFileName, ios::app);
		if (!LogFile.is_open())
		{
			perror("Couldn't create logfile. ");
			Logging = false;
			sleep(1);
		}
		else
			Logging = true;
	}
	if (Verbose)
	{
		snprintf(temp, sizeof(temp), "initializing a new KeySpaceManager with KeySpaceSizeInBits: %2.2x, Range %2.2x",
		KeySpaceSizeInBits, Range); 
		LogFile << temp;
	}
	
	
	//ncurses configuration
	#ifdef HAVE_NCURSES

//	initscr();
//	start_color();
//	init_pair(1, COLOR_YELLOW, COLOR_BLUE);
//	init_pair(2, COLOR_WHITE, COLOR_BLUE); 
	#endif

	NextInitialKey = new Key(KeySpaceSizeInBytes);
	srand(1);
	CheckedOutChunks.clear();
	StragglerList.clear(); //cant be to careful with STL stuff 

}

bool KeySpaceManager :: SaveState(FILE *fp)
{
		
	u_int32_t num_checkedout_chunks = getChunksCheckedOut();
	u_int32_t num_straggler_chunks = StragglerList.size();
	ChunkMapType :: iterator myIter;
	KeySpaceChunk *curr_chunk;
	#ifdef JC_DEBUG
	printf("Saving the state of KeySpaceManager\n");
	printf("writing out %d chunks\n", num_checkedout_chunks);
	#endif
	//write the number of checked out chunks followed by chunks
	fwrite(&num_checkedout_chunks, sizeof(num_checkedout_chunks), 1, fp);
	myIter = CheckedOutChunks.begin();
	for (int i = 0; i < num_checkedout_chunks; i++)
	{
		#ifdef JC_DEBUG
		printf("%d\t:", i);
		#endif
		curr_chunk =&(myIter->second);	
		curr_chunk->serialize(fp);
		myIter++;
	}
	// Save next initial key
	#ifdef JC_DEBUG
	cout << "Saving NextInitialKey: " << *NextInitialKey << endl;
	#endif
	NextInitialKey->serialize(fp);
	
	//save all the statistics
	fwrite(&CanComputeIntegerStats, sizeof(CanComputeIntegerStats), 1, fp);
	fwrite(&IntTotalChunks, sizeof(IntTotalChunks), 1, fp);
	fwrite(&IntCheckedOutChunkCounter, sizeof(IntCheckedOutChunkCounter), 1, fp);
	fwrite(&IntCheckedInChunkCounter, sizeof(IntCheckedInChunkCounter), 1, fp);
	fwrite(&IntExpiredCounter, sizeof(IntExpiredCounter), 1, fp);
	if (ferror(fp))
	{
		fprintf(stderr, "Error writing KeySpaceManager state info.\n");
		perror("fwrite:");
		return false;
		//exit(0);
	}
	CheckedOutChunkCounter.serialize(fp);
	CheckedInChunkCounter.serialize(fp);
	//save all the time info
	TotalComputeTime.serialize(fp);
	TotalRunningTime.serialize(fp);	
	
	#ifndef HAVE_NCURSES
	PrintStatistics();
	#endif
}

bool KeySpaceManager :: RestoreState(FILE *fp)
{
	u_int32_t num_checkedout_chunks;
	KeySpaceChunk curr_chunk;
	Key k;
	int i = 0;
	//read in the checked out chunks
	printf("restoring the state of KeySpaceManager\n");
	fread(&num_checkedout_chunks, sizeof(num_checkedout_chunks), 1, fp);
	printf("reading in %d chunks\n", num_checkedout_chunks);
	for (i = 0; i < num_checkedout_chunks; i++)
	{
		printf("%d\t:", i);
		curr_chunk = unserializeKeySpaceChunk(fp);
		CheckedOutChunks[curr_chunk.iKey] = curr_chunk;
	}
	LookForStragglers();
	//read in saved initial key
	*NextInitialKey = unserializeKey(fp);
	cout << "Restoring NextInitialKey: " << *NextInitialKey << endl;

	//restore all the statistics
	fread(&CanComputeIntegerStats, sizeof(CanComputeIntegerStats), 1, fp);
	fread(&IntTotalChunks, sizeof(IntTotalChunks), 1, fp);
	fread(&IntCheckedOutChunkCounter, sizeof(IntCheckedOutChunkCounter), 1, fp);
	fread(&IntCheckedInChunkCounter, sizeof(IntCheckedInChunkCounter), 1, fp);
	fread(&IntExpiredCounter, sizeof(IntExpiredCounter), 1, fp);
	if (ferror(fp))
	{
		fprintf(stderr, "Error readng KeySpaceManager state info.\n");
		perror("fread:");
		exit(0);
	}
	CheckedOutChunkCounter = unserializeKey(fp);
	CheckedInChunkCounter = unserializeKey(fp);
	//restore all the time info
	TotalComputeTime = unserializeTimePiece(fp);
	TotalRunningTime = unserializeTimePiece(fp);	
	PrintStatistics();
		
}
// This function looks for keyspacechunks that have been checked out and not returned in a reasonable amount of time
// any that are found are placed in StragglerList and will be re-checked out by getKeySpaceChunk on the next
// request. It should probably be called every 15 minutes or so.
void KeySpaceManager :: LookForStragglers()
{
	ChunkMapType :: iterator myIter;
	list<KeySpaceChunk *>::iterator listIter;
	struct timeval now;
	KeySpaceChunk *curr_chunk;

	gettimeofday(&now, NULL);
	int count = 0;
	myIter = CheckedOutChunks.begin();
	while (myIter !=CheckedOutChunks.end())
	{
		//cout <<"------" << count << "---------\n";
		curr_chunk= &(myIter->second);
		//printf("looking at Chunk with sessionId %x", curr_chunk->session_id);
		//cout << "And key value: " << (curr_chunk->iKey) << endl;
		if ( now.tv_sec - curr_chunk->out_time.tv_sec >  StragglerThreshold)
		{
				//cout << "Found A Straggler.\n";
				bool dupe = false;
				listIter = StragglerList.begin();
				while (listIter != StragglerList.end())
				{
					if ((*listIter)->session_id == curr_chunk->session_id && 
						(*listIter)->iKey == curr_chunk->iKey) // already in list
					{
						//cout << "Skipping, already fond it,\n";
						dupe = true;
						break;
					}
						
				listIter++;	
				}
				if (!dupe)
				{
					//cout << "Addning it\n";	
					StragglerList.push_back(curr_chunk);
				}
		}
		//cout <<"------" << count << "---------\n";
		myIter++;
	}		
#ifdef JC_DEBUG
	cout << "DONE LOOKING FOR STRAGLLERS\n";
#endif

}
// While this function only -returns- a KeySpaceChunk, it creates a KeySpaceChunkSTatusType object and
// stores it into CheckedOutChunks
void KeySpaceManager ::  PrintStatistics()
{
		float percent_out, percent_in;
		char temp[1024];

		cout << "There are: " << getChunksCheckedOut() << "Chunks checked out Currently\n";
		cout << "We have checked out " << CheckedOutChunkCounter << " out of " << TotalChunks << endl;
		cout << "We have checked in " << CheckedInChunkCounter << " out of " << TotalChunks << endl;
		if (CanComputeIntegerStats)
		{
			percent_out = 100.0 *(double) IntCheckedOutChunkCounter / IntTotalChunks * 1.0;
			percent_in = 100.0 * (double) IntCheckedInChunkCounter / IntTotalChunks * 1.0;

			printf("percent in = %d /%d = %f\n\n", IntCheckedInChunkCounter, IntTotalChunks, percent_in);
			cout << "We have checked out " << IntCheckedOutChunkCounter 
					<< "out of " << IntTotalChunks << " TotalChunks (";
			printf("%2.1f)\n", percent_out);

			cout << "We have checked in " << IntCheckedInChunkCounter 
					<< "out of " << IntTotalChunks << " TotalChunks (";
			printf("%2.1f)\n", percent_in);
			
		}	
		cout << "TotalComputeTime: " << TotalComputeTime << endl;
}
bool KeySpaceManager :: getKeySpaceChunk(KeySpaceChunk &chunk)
{
	//KeySpaceChunk chunk;
	
	#ifdef JC_DEBUG
	cout << "KeySpaceManager::getKeyChunk\n";
	#endif
	if ( CheckedOutChunkCounter  == TotalChunks) // notihng left to assign.
			return false;
	

	if (StragglerList.size() != 0)
	{
		if (Verbose)
			LogFile << " Handing out a straggler\n";

		chunk = *StragglerList.front();
		gettimeofday(&(chunk.out_time), NULL);
		//this is just to update the checked out time value so it doesnt keep getting
		//flagged
		CheckedOutChunks[chunk.iKey] = chunk;
		StragglerList.pop_front(); // we remove it when its checked out or in, whichever first
		// now we've assigned this chunk out atleat twice. it keeps the
		// same session_id so anyone who has it can submit it.
		return true;
	}
		
		
	
	if (Verbose)
		LogFile << "##creating a new KeyChunk with iKey" << *NextInitialKey << endl;

	
	chunk.range = Range;
	chunk.session_id = rand();
	chunk.iKey = *NextInitialKey;
	gettimeofday(&(chunk.out_time), NULL);
	
	CheckedOutChunks[chunk.iKey] = chunk;
	CheckedOutChunkCounter++; // Key, Used for display. Not Limited to 32 bits

	if (CanComputeIntegerStats)
		IntCheckedOutChunkCounter++;// integer used for computing percent done and ETA. limited to 32 bits. 

	//NextInitialKey is the next initalvalue for the next keyspace chunk we send out
	*NextInitialKey += Range;

	#ifdef JC_DEBUG
	cout << "End KeySpaceManager::getKeyChunk\n";
	#endif
	return true;
}

u_int32_t KeySpaceManager :: getChunksCheckedOut()
{
	return CheckedOutChunks.size();
}

void KeySpaceManager :: submitCompletedKeySpaceChunk(Key iKey, u_int32_t session_id)
{
	
	
	// apparently 'count' means to check for..
	if (!CheckedOutChunks.count(iKey))
	{
		LogFile << "Malicious submission attempt. Key never sent out\n";
		return;
	}

	if (CheckedOutChunks[iKey].session_id != session_id)
	{
		LogFile << "Malicious submission attempt. Invalid session id.\n";
		return;
	}

	if (CheckedOutChunks[iKey].session_id == session_id)
	{
		if (Verbose)
			LogFile << "Correctly submitted KeySpaceChunk /w iKey # " << iKey << endl;
		//if this was a straggler we need to take it out of the list
		
		list<KeySpaceChunk *> ::iterator myIter;
		myIter = StragglerList.begin();
		while (myIter != StragglerList.end())
		{
			if (  (*myIter)->session_id == session_id && (*myIter)->iKey == iKey )
			{
				#ifdef JC_DEBUG
				cout << "just erased straggler.\n";
				#endif
				StragglerList.erase(myIter);
			break;
			}
			else
				myIter++;
		}
		gettimeofday(  &CheckedOutChunks[iKey].in_time, NULL);
		CheckedOutChunks[iKey].compute_time =
			   	CheckedOutChunks[iKey].in_time.tv_sec - CheckedOutChunks[iKey].out_time.tv_sec;
		CheckedInChunkCounter++; // Key, Used for display. Not Limited to 32 bits

		if (CanComputeIntegerStats)
			IntCheckedInChunkCounter++;// integer used for computing

		TotalComputeTime.add_secs(CheckedOutChunks[iKey].compute_time);

		CheckedOutChunks.erase(iKey);
	}

}	

void KeySpaceManager :: UpdateDisplay(u_int32_t num_secs)
{
	TotalRunningTime.add_secs(num_secs);
#ifdef HAVE_NCURSES
	RedrawDisplay();
#else
	PrintStatistics();
#endif
}


KeySpaceManager :: ~KeySpaceManager()
{
	#ifdef JC_DEBUG
	cout << "KeySpaceManager::~KeySpaceManager :: begin\n";
	#endif
	delete NextInitialKey;

	#ifdef JC_DEBUG
	cout << "KeySpaceManager::~KeySpaceManager :: end\n";
	#endif


	#ifdef HAVE_NCURSES
	endwin();
	#endif
}
void KeySpaceManager :: IncrementStragglerThreshold(u_int32_t value)
{
	GlobalArguments.straggler_threshold += value;
	StragglerThreshold = GlobalArguments.straggler_threshold;
}
void KeySpaceManager :: DecrementStragglerThreshold(u_int32_t value)
{
	GlobalArguments.straggler_threshold -= value;
	StragglerThreshold = GlobalArguments.straggler_threshold;
}
bool KeySpaceManager :: StillMoreChunks()
{
		if (CheckedInChunkCounter == TotalChunks)
		{	
			cout << "Checked in chunk counter == total cuhnks\n";
			cout << CheckedInChunkCounter << TotalChunks << endl;
		
				return false;
		}
		else
				return true;
}
	


//------ Everything below this line has to deal with the ncurses display--------

#ifdef HAVE_NCURSES
void KeySpaceManager :: RedrawDisplay()
{
	//initscr();
	
	init_pair(1, COLOR_YELLOW, COLOR_BLUE);
	init_pair(2, COLOR_WHITE, COLOR_BLUE); 

	attron(COLOR_PAIR(2));
	DrawBg();
	attron(COLOR_PAIR(1));
	DrawFg();
	refresh();
	refresh();

//	endwin();
}

void KeySpaceManager :: DrawFg()
{
	char temp[256];
	int i =72;

	double percent_in = 0;
	double percent_out = 0;
	TimePiece straggler_thresh; // not in class cause its only for display.
	straggler_thresh.add_secs(GlobalArguments.straggler_threshold);
	//cout << "STraggler Threshold::: " << s << endl;	

	if (CanComputeIntegerStats)
	{	
		percent_out = 100.0 *(double) IntCheckedOutChunkCounter / IntTotalChunks * 1.0;
		percent_in = 100.0 * (double) IntCheckedInChunkCounter / IntTotalChunks * 1.0;
	}
	//network bssid
	snprintf(temp, sizeof(temp), "%s",  GlobalArguments.BaseName);
	mvprintw(1,11, temp);	

	//size of keyspace
	snprintf(temp, sizeof(temp), "%2d", KeySpaceSizeInBits);
	mvprintw(1,42, temp);	
	// Status, running unless all keychunks are assigned, then waiting
	if (CheckedOutChunkCounter == TotalChunks)
		snprintf(temp, sizeof(temp), "%s", "Waiting");
	else
		snprintf(temp, sizeof(temp), "%s", "Running");
	mvprintw(1,55, temp);	
	
	//total run time
	snprintf(temp, sizeof(temp), "%s", TotalRunningTime.getString().c_str());
	mvprintw(3,18, temp);	
	// total compute time
	snprintf(temp, sizeof(temp), "%s", TotalComputeTime.getString().c_str());
	mvprintw(3,53, temp);	

	//ChunkSize
	snprintf(temp, sizeof(temp), "%d", GlobalArguments.chunksize_in_bits);
	mvprintw(4,13, temp);	
	//Currently checked out chunks
	snprintf(temp, sizeof(temp), "%d", getChunksCheckedOut());
	mvprintw(4,40, temp);	
	//CurrentStragglers
	snprintf(temp, sizeof(temp), "%d", StragglerList.size());
	mvprintw(4,67, temp);	
	
	//Percent complete
	if (CanComputeIntegerStats)
		snprintf(temp, sizeof(temp), "%2.4f", percent_in);
	else
		snprintf(temp, sizeof(temp), "??%");
	mvprintw(5,21, temp);	

	//Straggler threshold:
	snprintf(temp, sizeof(temp), "%s", straggler_thresh.getString().c_str());
	mvprintw(5,52, temp);	

	//Next  iKEy
	snprintf(temp, sizeof(temp), NextInitialKey->getString().c_str());
	mvprintw(7, 13, temp);	

	//Total number of keychunks
	snprintf(temp, sizeof(temp), "%s", TotalChunks.getString().c_str());
	mvprintw(9, 25, temp);	

	//Total checked out key chunks
	snprintf(temp, sizeof(temp), "%s", CheckedOutChunkCounter.getString().c_str());
	mvprintw(10, 25, temp);	

	//Total checked in key chunks
	snprintf(temp, sizeof(temp), "%s", CheckedInChunkCounter.getString().c_str());
	mvprintw(11, 25, temp);	
}

void KeySpaceManager :: DrawBg()
{
	
	int WIDTH, HEIGHT;
	int mode = 128;
	int i = 0;
	int j = 0;
	char title[256];
	snprintf(title, sizeof(title), "jc-wepcrack %s by Johnny Cache", VERSION_STRING);
	//draw spaces to get a solid background.
	if (mode == 40)
	{	
		WIDTH = 60;
		HEIGHT = 8;
	for (i = 0; i < WIDTH; i++)
			for (j = 0; j < HEIGHT; j++)
					mvprintw(j,i, " "); 
	
	for (i = 0; i < HEIGHT; i++)
			mvprintw(i, WIDTH - 1, "|");
	for (i = 0; i < WIDTH; i++)
		mvprintw(0,i,"-");
	
	i = (WIDTH  - strlen(title)) / 2;
	mvprintw(0,i, title);
	//mvprintw(0,0, "--------------jc-wepcrack 0.10 by Johnny Cache-------------");	
	mvprintw(1,0,"| Netwxrk: 00:00:00:00:00:00     Size: 00    ");
	mvprintw(2,0, "-----------------------------------------------------------");	
	mvprintw(3,0,"| Percent Complete:              Percent Checked out:    ");
	mvprintw(4,0,"| Next Key:  ");
	mvprintw(5,0,"| Last Completed Key:               ");
	mvprintw(6,0,"| Conesecutive Block:               ");
	mvprintw(7,0, "------------------------------------------------------------");	

	}
	else 
	{
	WIDTH = 71;
	HEIGHT = 13;
	for (i = 0; i < WIDTH; i++)
		for (j = 0; j < HEIGHT; j++)
			{
			mvprintw(j,i, " "); 
			}


	for (i = 0; i < HEIGHT; i++)
			mvprintw(i, WIDTH - 1, "|");
	
	for (i = 0; i < WIDTH; i++)
		mvprintw(0,i,"-");

	i = (WIDTH  - strlen(title)) / 2;
	mvprintw(0,i, title);

	mvprintw(1,0,"| Network: 00:00:00:00:00:00     KeySize: 00    Status :  "); 
	mvprintw(2,0, "|---------------------------------------------------------------------");	
	mvprintw(3,0,"| Total Run Time:                Total Compute Time:              ");
	mvprintw(4,0,"| Chunksize:      Chunks currently out:        Current Stragglers:  ");
	mvprintw(5,0,"| Percent Complete:           Straggler Threshold:               ");
	mvprintw(6,0, "|---------------------------------------------------------------------");	
	mvprintw(7,0,"| Next iKey:                                                      ");
	mvprintw(8,0, "|---------------------------------------------------------------------");	
	mvprintw(9,0,"| Total KeyChunks: ");
	mvprintw(10,0,"| KeyChunks checked out:    ");
	mvprintw(11,0,"| KeyChunks checked in:             ");
	mvprintw(12,0, "|---------------------------------------------------------------------");	



	}	
	
}
#endif
