#include <iostream>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <unistd.h>
#include "config.h"
#include "main.h"
#include "key.h"
#include "keyspacemanager.h"
#include "misc.h"
#include "print.h" 
#include "packet80211.h"
#include "decrypt.h"
#include "test.h"
#include "arguments.h"
#include "settings.h"
#include <list>
#include <signal.h> // on linux systems this is needed
#include "server.h"
#define DEBUG 1

//------------------------Important Global vars----------------
Packet_80211 *GlobalPacketToCrack;
Packet_80211 *GlobalPacketToVerify;
KeySpaceManager *GlobalKeySpaceManager;
ArgumentsType   GlobalArguments;
//------------------------Prototypes----------------------------
void usage();
void StartCracking();
void parse_args(int argc, char **argv);
void exit_cleanly(int sigtype);
void save_state(int sigtype);
void load_state(char *filename);

void parse_args(int argc, char **argv)
{

	char ch;
	u_int32_t tmp;
	bool k_flag = false;
	bool p_flag = false;
	bool f_flag = false;
	bool c_flag = false;
	bool s_flag = false;
	GlobalArguments.verbose = false;
	GlobalArguments.logging  = false;
	memset((void *) &GlobalArguments, 0, sizeof(GlobalArguments));

	while ((ch = getopt(argc, argv, "vf:l:k:c:p:s:")) != -1)
	switch (ch) 
	{
		case 'v': //verbose
			GlobalArguments.verbose = true;
		break;

		case 'k': //keysize
			k_flag = true;
			GlobalArguments.keysize_in_bits = atoi(optarg);
		break;
		case 'p': //port
			p_flag = true;
			GlobalArguments.port = atoi(optarg);
			break;

		case 'f': //packetfile
		f_flag = true;
		snprintf(GlobalArguments.PacketFileName, sizeof(GlobalArguments.PacketFileName), "%s", optarg);
		break;

		case 'l': // logfile
		GlobalArguments.logging = true;
		snprintf(GlobalArguments.LogFileName, sizeof(GlobalArguments.LogFileName), "%s", optarg);
		break;
		case 'c': //chunksize
		c_flag = true;
		GlobalArguments.chunksize_in_bits = atoi(optarg);	
		fprintf(stderr, "setting chunksize to %d\n", GlobalArguments.chunksize_in_bits);

		break;
		case 's': //straggler threshold
		s_flag = true;
		GlobalArguments.straggler_threshold = atoi(optarg);	
		break;

		case '?':
		default:
		{
			usage();
			exit(0);
		}
	}

	if (!s_flag)
			GlobalArguments.straggler_threshold = DEFAULT_STRAGGLER_THRESHOLD;
	else if (GlobalArguments.straggler_threshold <= 0)
		{
			fprintf(stderr, "Invalid straggler threshold (%d)\n", GlobalArguments.straggler_threshold);
			usage();
			exit(0);
		}
	if (!p_flag)
		GlobalArguments.port = DEFAULT_PORT;
	else if (GlobalArguments.port <= 0)
		{
			fprintf(stderr, "Invalid Port. %d\n", GlobalArguments.port);
			usage();
			exit(0);
		}


	if (!c_flag)
		{
			GlobalArguments.chunksize_in_bits = DEFAULT_CHUNKSIZE_IN_BITS;
		}
	else if(GlobalArguments.chunksize_in_bits <= 0 || GlobalArguments.chunksize_in_bits > 32)
		{

				cout << "using custom chunk size: " << GlobalArguments.chunksize_in_bits << endl;
			fprintf(stderr, "Invalid chunk size. %d\n", GlobalArguments.chunksize_in_bits);
			usage();
			exit(0);	
		}

	if (!f_flag)
	{
		usage();
		fprintf(stderr, "need to specify a pcap file\n");
		exit(0);
	}
	if (!k_flag)
	{
		usage();
		fprintf(stderr, "need to specify a key size\n");
		exit(0);
	}

	if (GlobalArguments.keysize_in_bits <= 0)
	{
		fprintf(stderr,"Invalid key size %d\n", GlobalArguments.keysize);
		exit(0);
	}
	
	if (GlobalArguments.keysize_in_bits % 8 != 0)
	{
		fprintf(stderr,"Invalid key size. Use bits. \n");
		exit(0);
	}
		
	if (GlobalArguments.keysize_in_bits > MAX_KEY_SIZE_IN_BITS)
	{
		fprintf(stderr, "Key size to big. If you really mean to use it edit libjcwepcrack/config.h\n");
		exit(0);
	}

	GlobalArguments.keysize = GlobalArguments.keysize_in_bits / 8;
}
int main (int argc, char * argv[]) 
{
	// get arguments from state file
	if (argc == 1)
	{
		usage();
		exit(0);
	}
	if (argc == 2) // loading saved state file
	{
		load_state(argv[1]);
	}
	//or from the cmd line
	else
	{
		parse_args(argc, argv);
		getBssIdToCrack(GlobalArguments.PacketFileName); // fills in GlobalArguments.bbsidtocrack
		snprintf(GlobalArguments.BaseName, sizeof(GlobalArguments.BaseName), "%2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x", 
		GlobalArguments.BssIdToCrack[0], GlobalArguments.BssIdToCrack[1], GlobalArguments.BssIdToCrack[2], 
		GlobalArguments.BssIdToCrack[3], GlobalArguments.BssIdToCrack[4], GlobalArguments.BssIdToCrack[5]);
		GlobalKeySpaceManager = new KeySpaceManager(GlobalArguments.keysize_in_bits, GlobalArguments.chunksize_in_bits);
	}	

	GlobalPacketToCrack = getsmallpkt(GlobalArguments.PacketFileName); 
	GlobalPacketToVerify = getverifypkt(GlobalArguments.PacketFileName);

	
//	signal(SIGINT, exit_cleanly);
//	signal(SIGUSR1, save_state);
	signal(SIGPIPE, SIG_IGN);
	/*
	if ( testEverything() )
	{
		cout << "Good. All Tests PAssed\n";
		sleep(3);
	}
	else
	{
		cout << "Bad. FAiled a test\n";
		exit(0);
	}

*/
	printf("cracking network: ");
	print_hex(GlobalArguments.BssIdToCrack, sizeof(GlobalArguments.BssIdToCrack));
	
	printf("Using Packet: \n");
	GlobalPacketToCrack->print();
	u_char_t *packet_data = new u_char_t[GlobalPacketToCrack->getDataLength()];
	GlobalPacketToCrack->getData(packet_data);
	cout << "----------\n";
	print_hex(packet_data, GlobalPacketToCrack->getDataLength());
	cout << "----------\n";
	
	
	//testTimePiece();
//	testKeyClass();
	//testKeyChunk();
	//testKeyMap();
	//exit(0);
	//testKeySpaceManager();
	//GlobalKeySpaceManager->UpdateDisplay();
	//sleep(5);
	init_server();
	process_connections();
//	printf("=============================Start Cracking=========================\n");
	//StartCracking();
	delete GlobalKeySpaceManager;
	cout << "Exiting program\n";
	return 0;
}



void StartCracking()
{
	Key k(GlobalArguments.keysize);
	u_int32_t encrypted_data_size;
	u_int8_t *encrypted_data;
	u_int8_t encrypt_iv[3];
	
	u_int32_t verify_data_size;
	u_int8_t *verify_data;;
	u_int8_t verify_iv[3];

	bool success_decrypt = false;
	bool success_verify = false;
	
	encrypted_data_size = GlobalPacketToCrack->getDataLength();
	encrypted_data = new u_int8_t[encrypted_data_size];
	GlobalPacketToCrack->getWepIV(encrypt_iv);
	GlobalPacketToCrack->getData(encrypted_data);	

	verify_data_size = GlobalPacketToVerify->getDataLength();	
	verify_data = new u_int8_t[verify_data_size];
	GlobalPacketToVerify->getWepIV(verify_iv);
	GlobalPacketToVerify->getData(verify_data);

	/*
	k.value[0] = 0x00;
	k.value[1] = 0x11;
	k.value[2] = 0x22;
	k.value[3] = 0x33;
	k.value[4] = 0x30;
*/
	cout << "Attempting to crack a " << k.getSize() << "byte key\n";
	
	success_decrypt = DecryptWepData(encrypted_data, encrypted_data_size, encrypt_iv, k.value, k.getSize());
	if (success_decrypt) // double check /w another packet
		success_verify= DecryptWepData(verify_data, verify_data_size, verify_iv, k.value, k.getSize());
	if(success_verify)
	{
		printf("Success.\n");
		printf("network:\t");
		print_hex((u_int8_t*) &GlobalArguments.BssIdToCrack,6); 
		printf("Key:\t");
		cout << k << endl;
		exit(0);
	}

}
	
	

void usage()
{
	printf("jc-wepcrack by Johnny Cache. A nologin.org production. Version %s\n",
	VERSION_STRING);
	//printf("./jc-wepcrack [ -v (verbose) -c chunksize -l logfilename  -p port\n");
	printf("./jc-wepcrackd [-v -c chunksize -p port -s straggle_threshold] -f packetfile -k keysize\n\n");
	printf("-v Verbose flag. Doesn't change much really. \n");
	//printf("-v Verbose flag. Applies to output as well as logging\n");
	printf("-c chunksize. size of chunk for clients to compute in bits Default: %d \n", DEFAULT_CHUNKSIZE_IN_BITS);
	//printf("-l Logfile. If none specified...\n");
	printf("-p port. Port to listen on. Default: %d\n", DEFAULT_PORT);
	printf("-s straggle_threshold. Default: %d\n", DEFAULT_STRAGGLER_THRESHOLD);
	printf("-f packetfile. pcap file to crack\n");
	printf("-k keysize. The size of the key. Usually 40, 104 is also widely implemented.\n\n");

}
void save_state(int sigtype)
{
		FILE *fp;
		char temp[256];
		int rc;
		snprintf(temp, sizeof(temp), "%s.state", GlobalArguments.BaseName);
		fp = fopen(temp, "w+");
		if (!fp)
		{
			perror("Could not open state file for saving!\n");
		}
		rc = fwrite(&GlobalArguments, sizeof(GlobalArguments), 1, fp);
		if (rc != 1)
			perror("Error writing state file!\n");

		GlobalKeySpaceManager->SaveState(fp);
		fclose(fp);
}
void load_state(char *filename)
{
		FILE *fp;
		char temp[256];
		int rc;
		snprintf(temp, sizeof(temp), "%s", filename);
		printf("restoring state from file %s\n", temp);
		fp = fopen(temp, "r");
		if (!fp)
		{
			perror("Could not open state file for reading!\n");
			exit(0);
		}
		rc = fread(&GlobalArguments, sizeof(GlobalArguments), 1, fp);
		if (rc != 1)
			perror("Error reading state file!\n");
		GlobalKeySpaceManager = new KeySpaceManager(GlobalArguments.keysize_in_bits, GlobalArguments.chunksize_in_bits);
		GlobalKeySpaceManager->RestoreState(fp);
		GlobalKeySpaceManager->LookForStragglers();
		GlobalKeySpaceManager->PrintStatistics();
		fclose(fp);
}


void exit_cleanly(int sigtype)
{
#ifdef JC_DEBUG
	cout << "EXiting cleanly\n";
#endif
	#ifdef HAVE_NCURSES
	endwin();
	#endif
	save_state(0);
	delete GlobalPacketToCrack;
	delete GlobalPacketToVerify;
	exit(0);
}
