#include "server.h"
#include "settings.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>
//#include "datatypes.h"
#include "keyspacemanager.h"
#include "jcwepcrack.h"
#define BACKLOG 10     // how many pending connections queue will hold


void init_server()
{
	#ifdef HAVE_NCURSES
	initscr();
	start_color();
	#endif

}
void  process_connections()
{

     int sockfd, new_fd;  // listen on sock_fd, new connection on new_fd
     struct sockaddr_in my_addr;    // my address information
     struct sockaddr_in their_addr; // connector's address information
     int sin_size;
     int yes=1;

     if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) 
	 {
     	perror("socket");
     	exit(1);
     }

     if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int)) == -1) 
	 {
      	perror("setsockopt");
      	exit(1);
      }
        
     my_addr.sin_family = AF_INET;         				// host byte order
     my_addr.sin_port = htons(GlobalArguments.port);     // short, network byte order
     my_addr.sin_addr.s_addr = INADDR_ANY; 				// automatically fill with my IP
     memset(&(my_addr.sin_zero), '\0', 8); 				// zero the rest of the struct

     if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) == -1) 
	 {
    	perror("bind");
       	exit(1);
     }

     if (listen(sockfd, BACKLOG) == -1) 
	 {
     	perror("listen");
      	exit(1);
     }

	int strag_counter = 0;
	int save_counter = 0;
	while(GlobalKeySpaceManager->StillMoreChunks())
	{  // main accept() loop
    	sin_size = sizeof(struct sockaddr_in);
		bool ret;
		
		struct timeval tv;
		fd_set readfds;

		tv.tv_sec = 5;
		tv.tv_usec = 0;

		FD_ZERO(&readfds);
		FD_SET(sockfd , &readfds);
		//#ifdef HAVE_NCURSES
		FD_SET(0, &readfds);//stdin
		//#endif
  		// don't care about writefds and exceptfds:
		signal(SIGINT, exit_cleanly);
		signal(SIGUSR1, save_state);

		struct timeval before;
		struct timeval after;
		int waited_secs;
		
		gettimeofday(&before, NULL);
  		select(sockfd+1, &readfds, NULL, NULL, &tv);
		gettimeofday(&after, NULL);
		waited_secs = after.tv_sec - before.tv_sec;

		signal(SIGINT, SIG_IGN);
		signal(SIGUSR1, SIG_IGN);

		GlobalKeySpaceManager->UpdateDisplay(waited_secs);

		strag_counter += waited_secs;
		save_counter += waited_secs;
		if (strag_counter >= GlobalArguments.straggler_threshold)
		{
			GlobalKeySpaceManager->LookForStragglers();
			strag_counter = 0;
		}
		if (save_counter >= FIFTEEN_MINS)	
		{
			GlobalKeySpaceManager->LookForStragglers();
			save_state(0); // saves everything.
			save_counter = 0;
		}
		
  		//
		if (FD_ISSET(0, &readfds))
		{
			char ch;
			int r;
			r = read(0, (char *) &ch, sizeof(ch));
			if (r > 0)
			{
				if (ch =='q')
					exit_cleanly(0);
				if (ch == '-')
					GlobalKeySpaceManager->DecrementStragglerThreshold();
				else if (ch == '+')
					GlobalKeySpaceManager->IncrementStragglerThreshold();
				else if (ch =='[')
					GlobalKeySpaceManager->DecrementStragglerThreshold(60);	
				else if (ch ==']')
					GlobalKeySpaceManager->IncrementStragglerThreshold(60);	
				
			}
				

		}
		

		
  		if (!FD_ISSET(sockfd, &readfds))
		{		
			continue;
		}
		else
		
        if ((new_fd = accept(sockfd, (struct sockaddr *)&their_addr,(socklen_t*)  &sin_size)) == -1) 
		{
        	perror("accept"); 
			continue;
        }
		
#ifdef JC_DEBUG
        printf("server: got connection from %s\n", inet_ntoa(their_addr.sin_addr));
#endif
		
		struct ProtocolHeader h;
		memset((void *) &h, 0, sizeof(h));

		u_int8_t *data;
		ret = read_packet(new_fd, h, &data);

		if (!ret)
			continue;

		if (h.size  == 0)
		{
			data =  NULL;
#ifdef JC_DEBUG
			cout <<  "READ 0 SIZED PACKET\n";
#endif
		}
		else
	
		if (data == NULL)
		{
			cout << "DATA IS NULL\n";
			sleep(1);
		
		}
#ifdef JC_DEBUG
		printf("read in header. \n");
		print_header(h);
#endif
		if (h.major != MAJOR_PROTOCOL_VERSION)
		{
			printf("Protocol version mis match. dropping connection\n");
			close(new_fd);
			if (data)
				delete data;
			continue;
		}		
		if (h.RequestType == REQUEST_DATA)				
		{
#ifdef JC_DEBUG
			printf("Client wants new keyspace chunk\n");
#endif
			InitiateNewSession(new_fd);
			close(new_fd);
			if (data)
				delete data;

		}
		else if (h.RequestType == SUBMIT_COMPLETED_CHUNK)
		{
#ifdef JC_DEBUG
			printf("Client Submitting Completed chunk\n");
#endif
			close(new_fd);
			AcceptSubmittedChunk(h,data);
			if(data)
				delete data;
			// we need a smarter function for these, similliar to SessionData abstraction.
		}
		else if (h.RequestType == SUBMIT_VERIFIED_KEY)
		{
			printf("Client Submitting cracked key!\n");
			close(new_fd);
			AcceptSubmittedKey(h, data);
			if(data)
				delete data;

		}
		else 
		{
			printf("Invalid request type %x. closing connection\n", h.RequestType);
			close(new_fd);
			if(data)
				delete data;
		}

    }

	cout << "ALL DONE.\n Key was not cracked. Probably a different key size.\n";
	exit_cleanly(0);

	return;
} 

//This is called by the server to start anew session with the client. It gets teh data in order
//and calls libjcwepcrack to send it off
bool InitiateNewSession(int fd)
{

	bool ret;
	//ProtocolHeader h;
	KeySpaceChunk k;
	struct SessionData session_data; // from libjcwepcrack
	
	int rc;
	//Get the packet data in line
	//
    
    session_data.packet1size= GlobalPacketToCrack->getDataLength();
    session_data.packet1_data = new u_int8_t[session_data.packet1size];
    GlobalPacketToCrack->getWepIV(session_data.packet1_iv);
    GlobalPacketToCrack->getData(session_data.packet1_data);   

    session_data.packet2size = GlobalPacketToVerify->getDataLength();   
    session_data.packet2_data = new u_int8_t[session_data.packet2size];
    GlobalPacketToVerify->getWepIV(session_data.packet2_iv);
    GlobalPacketToVerify->getData(session_data.packet2_data);

	if (!session_data.packet1_data || !session_data.packet2_data)
	{
		cout << "couldnt allocate memory for packet. exiting\n";
		exit(0);
	}
	ret =  GlobalKeySpaceManager->getKeySpaceChunk(k);
	if (!ret)
	{
		cout << "Couldnt get a new keyspace chunk from keyspacemanager. Dropping connection\n";
		delete []session_data.packet1_data;
		delete []session_data.packet2_data;
		close(fd);
		return false;
	}
	session_data.myKeySpaceChunk = &k; // on stack, dont need to free.


	//HOWTO DEBUG THIS MEMORY LEAK:
	// SEE HOW MUCH MEMORY IS TAKEN UP AFTER 1000 CHUNKS ARE CHECKED OUT
	// -AND- SENT ACROSS THE NETWORK, 
	// VS HOW MUCH MEMORY IS TAKEN UP AFTER 1000 CHUNKS ARE CHECKED OUT 
	// AND NEVER SENT (COMMENT SendSessionData OUT);
	ret = SendSessionData(fd, session_data);
	delete []session_data.packet1_data;
	delete []session_data.packet2_data;
	close(fd);


	return ret;

}



bool AcceptSubmittedChunk(ProtocolHeader &h, u_int8_t *data)
{
	int rc;
	bool ret;
	u_int32_t session_id;
	Key iKey;


	// we consider the data to be opaque, we dont want to depends on
	// protocol conventions. let the library parse it
	ret = ParseFinishedKeyOrKeyChunk(h,data, session_id, iKey);
	if (ret)
	{
	#ifdef JC_DEBUG
		cout << "Received keyspacechunk submission with session_id " << session_id << " and iKey " << iKey;
	#endif
	 	GlobalKeySpaceManager->submitCompletedKeySpaceChunk(iKey, session_id);
	 	return true;
	}
	else
	{
	#ifdef JC_DEBUG
		cout << "Received keyspacechunk submission but couldnt read data. \n";
	#endif
	}	
}
bool AcceptSubmittedKey(ProtocolHeader &h, u_int8_t *data)
{
	int rc;
	bool ret;
	u_int32_t session_id;
	Key iKey;


	// we consider the data to be opaque, we dont want to depends on
	// protocol conventions. let the library parse it
	ret = ParseFinishedKeyOrKeyChunk(h, data, session_id, iKey);
	if (ret)
	{
	#ifdef JC_DEBUG
		cout << "Received cracked key submission with session_id " << session_id << " and iKey " << iKey;
	#endif
		if (VerifyPacket(iKey))
		{
			exit_cleanly(0);
		}
	}
	else
	#ifdef JC_DEBUG
		cout << "Received key submission but couldnt parse data. \n";
	#endif
	return false;
	
}






bool VerifyPacket(Key k)
{
	
	FILE *fp;
 	u_int32_t encrypted_data_size;
    u_int8_t *encrypted_data;
    u_int8_t encrypt_iv[3];

    u_int32_t verify_data_size;
    u_int8_t *verify_data;;
    u_int8_t verify_iv[3];

    bool success_decrypt = false;
    bool success_verify = false;
    
    encrypted_data_size = GlobalPacketToCrack->getDataLength();
    encrypted_data = new u_int8_t[encrypted_data_size];
    GlobalPacketToCrack->getWepIV(encrypt_iv);
    GlobalPacketToCrack->getData(encrypted_data);

    verify_data_size = GlobalPacketToVerify->getDataLength();
    verify_data = new u_int8_t[verify_data_size];
    GlobalPacketToVerify->getWepIV(verify_iv);
    GlobalPacketToVerify->getData(verify_data);

	success_decrypt = DecryptWepData(encrypted_data, encrypted_data_size, encrypt_iv, k.value, k.getSize());
   	if (success_decrypt) // double check /w another packet
   		success_verify= DecryptWepData(verify_data, verify_data_size, verify_iv, k.value, k.getSize());
        
	delete encrypted_data;
	delete verify_data;
    if(success_verify)
        {
            printf("Success.\n");
            printf("network: %s\n", GlobalArguments.BaseName);

			cout << "Key: " << k << endl;
			char tmp[256];
			snprintf(tmp, sizeof(tmp), "%s.KEY", GlobalArguments.BaseName);

			fp = fopen(tmp, "w+");
			if (!fp)
			{
				cout << "could not open file %s for logginf the key to.\n";
			}
			else
			{
				fprintf(fp, "network: %s\n", GlobalArguments.BaseName);
				fprintf(fp, "Key: %s\n", k.getString().c_str());
				fclose(fp);
			}
			
            exit(0);
        }
	else
	{
		cout << "False Alarm. \n";
		//sleep(2);
		return false;
	}
}
