#include <iostream>
#include <unistd.h>
#include "config.h"
#include "main.h"
#include "key.h"
#include "keyspacemanager.h"
#include "keyspacechunk.h"
#include "print.h"
#include "misc.h"
#include "packet80211.h"
#include "decrypt.h"
#include "datatypes.h"
#include "test.h"
#define DEBUG 1
map<Key, bool> TestMap;
extern KeySpaceManager * GlobalKeySpaceManager;


bool testEverything()
{
	if (testKeyClass() && testKeyMap() && testKeyChunk() && testTimePiece() && testKeySpaceManager())
	{
		cout << "good. all tests passed\n";
		return true;
	}
	else
	{
		cout << "Bad. a test failed. Please look through output.\n";
		exit(0);
		return false;
	}
}
		
bool testKeySpaceManager()
{
	cout << "Testing KeySpaceManager\n";

	KeySpaceChunk k1, k2, k3;

	KeySpaceChunk keys[100];
	/*
	for (int i  = 0; i < 100; i++)
		{
		GlobalKeySpaceManager->getKeySpaceChunk(keys[i]);
		GlobalKeySpaceManager->submitCompletedKeySpaceChunk(keys[i].iKey, keys[i].session_id);
		}
	GlobalKeySpaceManager->getKeySpaceChunk(k1);
	GlobalKeySpaceManager->getKeySpaceChunk(k2);
	GlobalKeySpaceManager->LookForStragglers();
	GlobalKeySpaceManager->PrintStatistics();
*/
	while (1)
	{
	GlobalKeySpaceManager->getKeySpaceChunk(k3);
	GlobalKeySpaceManager->submitCompletedKeySpaceChunk(k3.iKey, k3.session_id);
//	GlobalKeySpaceManager->UpdateDisplay();
	}

	
//	sleep(2);

	return true;
}
bool testKeyClass()
{
	Key k1(5);
	Key k2(5);
	Key k3;
	k1 += 1;
	k2 += 2;

	cout << "Testing key class\n";

	if (k1 < k2)
		printf("Key passes test!\n");
	else
	{
		printf("Key Class fails < test. :(\n");
		return false;
	}

	k3 = k1;
	k3++;
	cout << "k3 should be 1 more than k1\n";
	cout << "k1: " << k1 << "k3 " << k3 << endl;
	if (k1 < k3)
		cout << "Key class passes test\n";
	else
	{
		cout << "Key class fails increment test\n";
		return false;		
	}
	serialized_object ser_key;
	ser_key = k3.serialize();
	cout << "just serialized k3. it is " << ser_key.size <<" bytes" << endl;
	Key sKey = unserializeKey(ser_key);
	cout << "Just unserialized that same key it should == k3\n";
	delete ser_key.data;
	if (sKey == k3)
		{
		cout << "Key class passes test!\n";
		cout << k3 <<"==" << sKey << endl;
		}
	else
		{
		cout << "Key class failes serialize to memory test\n";
		return false;
		}

	FILE *fp;
	fp = fopen("./test.dat", "w+");
	if (!fp)
	{
		cout << "could not create file to test key class\n";
		return false;
	}

	k2.serialize(fp);
	cout << "just serialized k2 to test.dat\n";
	rewind(fp);
	Key k5 = unserializeKey(fp);
	cout << "just unserialized test.dat to k5. if k2 == k5 were good\n";
	if (k5 == k2)
		{
		cout << "Key class passes test!\n";
		cout << k3 <<"==" << sKey << endl;
		}
	else
		{
			cout << "Key class fails serialize to disk test.\n";
			return false;
		}
			
	cout << "Key class passed all tests\n";
	return true;	
}
bool testKeyChunk()
{
	FILE *fp;
	KeySpaceChunk chunk1, chunk2, chunk3;
	Key iKey;
	iKey += 100;
	cout << "Testing KeySpaceChunk\n";
	// chunk1 is the initial chunk
	// make chunk2 and chunk3 out of serialized versions of chunk1
	chunk1.iKey = iKey;
	chunk1.compute_time = 22;

	cout << "about to serialize KeySpaceChunk to memory\n";
	serialized_object ser_chunk;
	ser_chunk = chunk1.serialize();
	cout << "about to unserailize chunk back\n";
	chunk2 = unserializeKeySpaceChunk(ser_chunk);

	//serialze to memory test
	if (chunk2.compute_time == 22 && chunk2.iKey == iKey)
	{
		cout << "KeySpaceChunk passed test. Unserialized Chunk == chunk\n";
		cout << chunk1.iKey << "==" << chunk2.iKey << endl;
		cout << chunk1.compute_time << "== " <<  chunk2.compute_time << endl;
	}
	else
	{
		cout << "KeySpaceChunk failed serialize to memory test!\n";
		cout << chunk1.iKey << "==" << chunk2.iKey << endl;
		cout << chunk1.compute_time << "== " <<  chunk2.compute_time << endl;
		return false;
	}
	
	cout <<"---------------\n";	
	cout << "serializing chunk to file\n";
	fp = fopen("keychunk.dat", "w+");
	if (!fp)
	{
		cout << "couldnt create keychunk.dat to test KeySpaceChunk class\n";
		return false;
	}
	chunk1.serialize(fp);

	rewind(fp);
	cout <<"unserializing chunk\n";
	chunk3 = unserializeKeySpaceChunk(fp);
	cout << "if chunk3. compute_time = 22 and chunk3.iKey == iKey this workd\n";
	cout << "chunk3.compute_time = " << chunk3.compute_time;
	cout << "chunk3.iKey ==%d\n" <<  chunk3.iKey << endl;
	if (chunk3.compute_time == 22 && chunk3.iKey == iKey)
		cout << "KeySpaceChunk passed test!\n";
	else
	{
		cout <<"KeySpaceChunk failed test!\n";
		return false;
	}

	cout << "KeySpaceChunk passed all tests\n";
	return true;
}

bool testKeyMap()
{
	map<Key, bool>::iterator myIter;
	Key k1(5);
	Key k2(5);
	Key k3(5);
	k1 += 2;
	k2 += 1;
	
	cout << "testing KeyMap\n";
	TestMap[k1] = true;
	TestMap[k2] = true;	

	cout << "there should be 2 keys in the map. There are: " << TestMap.size() << "keys";
	if (TestMap.size() == 2)
		cout << "Test sucessful!\n";
	else
	{
		cout << "yikes. looks like the KeyMap class is really broken\n";
		return false;
	}

	k1++;
	TestMap[k1] = true;

	cout << "there should be 3 keys in the map. There are: " << TestMap.size() << "keys";
	if (TestMap.size() == 3)
		cout << "Test sucessful!\n";
	else
	{
		cout << "Test keymap failed. looks like memory allocation issues?\n";
		return false;
	}

	cout << "KeyMap passed all tests\n";
	return true;
}


bool testTimePiece()
{
	FILE *fp;
	TimePiece t1, t2;
	t1.add_secs(120);

	cout << "testing TimePiece\n";
	fp = fopen("timepiece.dat", "w+");
	if (!fp)
	{
		cout << "couldnt create timpeice.dat to test timepiece class\n";
		return false;
	}
	t1.serialize(fp);
	rewind(fp);
	t2 = unserializeTimePiece(fp);
	cout << "if t1 == t2, TimePiece.serialize works\n";
	cout << t1 << "==" << t2 << endl;	

	cout << "TimePiece passed all tests\n";
	return true;
}

