#include "timepiece.h"



TimePiece::TimePiece()
{
	day_counter = 0;
	hour_counter = 0;
	min_counter = 0;
	sec_counter = 0;
}

void TimePiece::add_secs(u_int32_t secs)
{
	sec_counter += secs;
	if (sec_counter >= 60)
	{
		min_counter += sec_counter / 60;
		sec_counter %= 60;
	}

	if (min_counter >= 60)
	{
		hour_counter += min_counter / 60;
		min_counter %= 60;
	}

	if (hour_counter >= 24)
	{
		day_counter += hour_counter / 24;
		hour_counter %= 24;
	}
}

void TimePiece :: reset()
{
	day_counter = 0;
	hour_counter = 0;
	min_counter = 0;
	sec_counter = 0;
}
bool TimePiece ::serialize(FILE *fp)
{
	fwrite( (void *) &day_counter, sizeof(day_counter),   1, fp);
	fwrite( (void *) &hour_counter, sizeof(hour_counter), 1, fp);
	fwrite( (void *) &min_counter, sizeof(min_counter),   1, fp);
	fwrite( (void *) &sec_counter, sizeof(sec_counter),   1, fp);
	if (ferror(fp))
	{
		cout << "Error serializing timepiece.\n";
		perror("fwrite:");
		exit(0);
		return false; //maybe dont want to exit sometime.
	}
//	printf ("wrote timepiece data to %d %d %d %d\n", day_counter, hour_counter, min_counter, sec_counter);

	return true;
}

TimePiece unserializeTimePiece(FILE *fp)
{
	TimePiece t;
	u_int32_t day, hour,min,sec;	
	fread( (void *) &day, sizeof(day),   1, fp);
	fread( (void *) &hour, sizeof(hour), 1, fp);
	fread( (void *) &min, sizeof(min),   1, fp);
	fread( (void *) &sec, sizeof(sec),   1, fp);
	if (ferror(fp))
	{
		cout << "Error unserializing timepiece.\n";
		perror("fread:");
		exit(0);
	}
//	printf ("Setting new timepiece data to %d %d %d %d\n", day, hour, min, sec);
	t.set_day(day);
	t.set_hour(hour);
	t.set_min(min);
	t.set_sec(sec);
//	cout << "After initialization: " << t;
	return t;
}

string TimePiece :: getString()
{
	char temp[1024]; // this out to be enough.

	sprintf(temp, "%dd %dh %dm %ds", day_counter, hour_counter, min_counter, sec_counter);
	string s = temp;
	return s;
	
}
//ostream &operator<<(ostream &o, const TimePiece &t)
ostream &operator<<(ostream &o, TimePiece &t)
{
	o << t.getString();
	return o;
}
