#include "packet80211.h"
#include "datatypes.h"
#include "print.h"
#include <netinet/in.h> //ntohs 

Packet_80211 :: Packet_80211(struct pcap_pkthdr p, u_char_t *data) : Packet(p, data)
{
	memcpy((void *) &wlan_header, (void *) data, sizeof(wlan_header));

}
void Packet_80211 :: print()
{
		print(cout);
}

void Packet_80211 :: print(ostream &s)
{
	u_int8_t src[6], dst[6];
	u_int8_t iv[3];
	u_int8_t keynum;
	u_int8_t icv[4]; //checksum
	
	getSrcAddr(src);
	getDstAddr(dst);
	getWepIV(iv);
	getWepKeyNum(&keynum);
	getWepICV(icv);
	char tmp[256];

	snprintf(tmp, sizeof(tmp), "[%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x]", src[0], src[1], src[2], src[3], src[4], src[5]);
	s << tmp;
	s << "==>";
	snprintf(tmp, sizeof(tmp), "[%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x]", dst[0], dst[1], dst[2], dst[3], dst[4], dst[5]);
 	s << tmp << endl;
	cout << "Wep: " << isWep() << endl;

	snprintf(tmp, sizeof(tmp), "%2.2x%2.2x%2.2x", iv[0], iv[1], iv[2]);
	cout << "IV: " << tmp << endl;

	snprintf(tmp, sizeof(tmp), "%2.2x", keynum);
	cout << "Keynumber (usually 0): " << tmp << endl;
	snprintf(tmp, sizeof(tmp), "%2.2x%2.2x%2.2x%2.2x", icv[0], icv[1], icv[2], icv[3]);
	cout << "ICV (checksum): "	<< tmp << endl;
	//Packet::print_hex_dump();
	
}


bool Packet_80211 :: isEntering()
{
	u_int16_t flag = 0x0001;
	#ifdef LITTLE_ENDIAN
	flag = htons(flag);
	#endif
	if (wlan_header.frame_ctrl & flag)
			return true;
	else
			return false;
}
bool Packet_80211 :: isExiting()
{
	u_int16_t flag = 0x0002;
	#ifdef LITTLE_ENDIAN
	flag = htons(flag);
	#endif
	if (wlan_header.frame_ctrl & flag)
		return true;
	else 
		return false;
}
bool Packet_80211 :: isWep()
{
	
	u_int16_t flag;
	flag = 0x0040;
	
	#ifdef LITTLE_ENDIAN
	flag = htons(flag);
	#endif
	if(wlan_header.frame_ctrl & flag)
		return true;
	else
		return false;
}
/*
bool Packet_80211 :: isNormal()
{
	print_hex((u_int8_t *) &wlan_header, 2);
	printf("\n");
	u_int8_t tmp = 0x80;
	if (memcmp((u_int8_t*) (&wlan_header.frame_ctrl), &tmp, 1) == 0)
		return true;
	else
		return true;
}
*/
void Packet_80211 :: getSrcAddr(u_int8_t *p)
{
	if (isEntering())
		memcpy(p,wlan_header.addr2, sizeof(wlan_header.addr2));
	else 
	if (isExiting())
		memcpy(p, wlan_header.addr3, sizeof(wlan_header.addr3));
}
void Packet_80211 :: getDstAddr(u_int8_t *p)
{
	if (isEntering())
		memcpy(p,wlan_header.addr3, sizeof(wlan_header.addr3));
	else 
	if (isExiting())
		memcpy(p, wlan_header.addr1, sizeof(wlan_header.addr1));
}
void Packet_80211 :: getBssId(u_int8_t *p)
{
	if (isEntering())
		memcpy(p,wlan_header.addr1, sizeof(wlan_header.addr1));
	else 
	if (isExiting())
		memcpy(p, wlan_header.addr2, sizeof(wlan_header.addr2));
}

bool Packet_80211 :: getWepIV(u_int8_t * p)
{
		if (!isWep())
		{
			p = NULL;
			return false;
		}

		memcpy(p, data +sizeof(wlan_header), 3); // 3 byte IV

		return true;
}
bool Packet_80211 :: getWepKeyNum(u_int8_t *p)
{
		if (!isWep())
		{
			p = NULL;
			return false;
		}

		memcpy(p, data +sizeof(wlan_header) + 3, 1); //directly after IV
		return true;
}

bool Packet_80211 :: getWepICV(u_int8_t *p)
{

		if (!isWep())
		{
			p = NULL;
			return false;
		}

		memcpy(p, data + length - 4, 4); //Last 4 bytes
		return true;

}

/*
 * Pay Attention here, we cant be off by a byte when it comes to this >:>.
 * if the packet is encrypted getDataLength returns the size of the PDU 
 * (thats IEEE talk for the data, PLUS 4 BYTES FOR THE ICV.
 * That is because the ICV is encrypted with the data and sent along.
 *
 * If its not encrypted then this is all straight forward
 */
// return size of the payload inside the 80211 header
int Packet_80211 :: getDataLength()
{
	if (isWep())
			return (length - sizeof(wlan_header) -4); // dont count IV.
	else
		return (length - sizeof(wlan_header));
}
void Packet_80211 :: getData(u_char_t *dst)
{
		if (isWep())							// +4 /-4 to account for not sending IV
			memcpy(dst, data + sizeof(wlan_header) + 4, length - sizeof(wlan_header) -4);
		else
			memcpy(dst,data + sizeof(wlan_header), length - sizeof(wlan_header));
}


